CON Touch
'' ILI9325 driver using the Touchblade161 design for faster ram to display transfers with a hc137 for group select
'' Average Joe and James Moxham 2012

' notes re groupings of pins
' To shut down a group, deselect it by making P22 low. This makes Y0-Y7 from the 137 all high
' To start a group, the initial state is P0-P20 all HiZ and P22 HiZ (P21 and 23 are audio and there is less audio noise if the pins are not next to each other)
' Passing to a cog, P0-P20 and P22 HiZ ie DIRA all as 0 fo these pins. This is to avoid contention between spin and the cog code
' A routine in Spin or Pasm that changes the 137 will also return P0-P20 and P22 as HiZ.
' In general terms, H is inactive and L is active. So changing a pin from HiZ to H will have no effect as pins all have pullups.
' Any handover to cog code must be done in this HiZ state and all pins in a group need to work with this


  Margin = 4                                            ' some characters have xoffset of -1 or -2 eg v, and won't display on the extreme left edge
  RamDiskStart = 0                                      ' start of the ram disk
  _1ms      = 1_000_000 /     1_000                         ' Divisor for 1 ms

  ' touchscreen pins
  Touch_Clock = 0                        
  'Touch_ChipSelect = done with a 137 select
  Touch_DataIn = 1            
  Touch_DataOut = 2' touchscreen pins

    ' kye sd numbers
  _dopin = 24
  _clkpin = 25
  _dipin = 26
  _cspin = 27
  _cdpin = -1 ' -1 if unused.
  _wppin = -1 ' -1 if unused.
  _rtcres1 = -1 ' -1 always.
  _rtcres2 = -1 ' -1 always.  
  _rtcres3 = -1 ' -1 always.  
  _statuspin = -1 ' Status LED pin number.

VAR
   byte FileNumber                                      ' 0 to 31
   byte DisplayMode                                     ' type of display and whether single or dual
   word ScreenWidth                                     ' either 240 in portrait or 320 in landscape
   word ScreenHeight                                    ' 320 in portrait or 240 in landscape
   long orientation                                     ' true is portrait
   long curx, cury     
   byte FontHeight              ' size of the current loaded font (pixels = fontsize *.75)
   word BackFontColor              ' the background color in RRRRRGGG GGGBBBBB format
   byte Latch                   ' current latch output   


OBJ
   spi:           "SPI_ASM"                               ' thanks to Beau - spi driver for touch screen    
   fat:           "SD-MMC_FATEngine.spin"                   ' Kye's latest sd driver and no RTC    
   str:             "ASCII0_STREngine.spin"                 ' thanks to Kye - string routines   
DAT 
    RamLocation         long $0[31]                  ' 32 file ram locatinos
    RamSize             long $0[31]                  ' File sizes in words - NOT bytes (divide byte by 2)
    sdbuffer            byte $0[1024] ' 1024 byte buffer for sd card interface and also for sending one row 320 x3 = 960 bytes max picture size 
    buffer2             byte $0[512]  ' 512 general purpose hub buffer 
    rambuffer           word $0[256]  ' 512 byte (256 word) buffer easier for using with words

PUB Begin
  start_ram_cog                                             ' start the external ram / display driver in a cog, needed for many routines so always first
  Change161(0)                                              ' reset all the 161 counters to zero
  SelectMemGroup                                            ' select group1
  Start_ILI9325(2)                                       ' start the display 0 = ILI9325 single, 1 = SSD1289 single, 2 ILI9325 double, 3, SSD1289 double                                       
  Text(string("SD"))                                    ' string to send, uses internal prop font so can see something if the SD fails to mount   
  SetOrientation(true)                                  '  in portrait (true) or landscape mode (false), must be before clearscreen otherwise don't know screensize   
  fat.fatEngineStart( _dopin, _clkpin, _dipin, _cspin, _wppin, _cdpin, _rtcres1, _rtcres2, _rtcres3)
  fat.mountPartition(0)                                 ' mount the sd card
  Clearscreen(RGBtoWord(0,0,0))                         ' clear screen to black - uses the display driver above
  RamErase(0)                                              ' erase ram chip fat
  LoadDesktop                                           ' a image 240x320 for the desktop. Load first as there is a ramclear routine that does not delete the first entry
  DrawDesktop

  'dira := 0                                            ' HiZ for debugging the SSD1289
  'repeat
  
  LoadIcons                                             ' load the mask and icons onto the screen
  SetCursor(0,0)                                        ' cursor for debugging to 0,0
  SelectSPIGroup                                        ' selects this group and starts the cog
                                  
PUB RamErase(n)
    Filenumber := n                                     ' erase lookup table for the ram files, effectively erases the files    
                                                        ' if 0 then erase all, if 1 then leave the desktop intact (assuming this was loaded first)

PUB GetCurX
    result := curx                                       ' get the text cursor position

PUB GetCurY
    result := cury

PUB SetCurx(n)                                            ' set the text cursor position
    Curx := n

PUB SetCury(n)
    Cury := n

PUB GetBackFontColor
   result := BackFontColor

PUB GetRamLocation(n)
   result := Ramlocation[n]

PUB GetFilenumber
   result := filenumber

PUB DrawDesktop
   DrawBMPRam(0,0,0)                                    ' file number zero, location 0,0

PUB LoadDesktop                                         ' first bmp file is not deleted wtih clearramdiskexcludedesktop
   SDBMPtoRam(string("balloon.bmp"))                     ' file number 0

PUB Chain(stringptr)
    result := \fat.bootpartition(stringptr) ' reboot and run this program            

PUB LoadIcons                                           ' load all the icons onto the screen. Possibly use a .ini text file for this
   SDBMPtoRam(string("mask.bmp"))                       ' file number 1
   MergeBackgroundIconRam(string("photos.bmp"),0,120)   ' file number 2 and leaves filenumber as 2 ready for the next one
   MergeBackgroundIconRam(string("calc.bmp"),60,120)        ' draw an icon onto the background    
   MergeBackgroundIconRam(string("facebook.bmp"),120,120)   ' draw an icon
   MergeBackgroundIconRam(string("videos.bmp"),180,120)     ' draw an icon
   MergeBackgroundIconRam(string("notes.bmp"),0,180)        ' draw an icon
   MergeBackgroundIconRam(string("music.bmp"),60,180)       ' draw an icon
   MergeBackgroundIconRam(string("ipod.bmp"),120,180)       ' draw an icon 
   MergeBackgroundIconRam(string("prop.bmp"),180,180)       ' prop icon for a GUI buttons, radio, checkboxes etc
   MergeBackgroundIconRam(string("itunes.bmp"),0,240)       ' play some music
   MergeBackgroundIconRam(string("safari.bmp"),60,240)       ' html (better GUI modelled on Opera rather than Windows)
   DrawDesktop

PUB SDtoRam(stringptr) | RamAddress,remainder,sizebytes,sizewords                              ' copy a file from the SD card to the ram disk
    OpenFileRead(stringptr)
    sizebytes := fat.filesize                           ' size in bytes
    sizewords := sizebytes >>1                          ' size in words
    ramaddress := NextLocation                          ' next free place in ram
    RamSize[FileNumber] := sizewords                    ' ram disk works in words, not bytes
    repeat sizebytes >> 9                               ' do blocks of 512 bytes each as natural size of SD blocks
      fat.readdata(@sdbuffer,512)                       ' read from SD card
      docmd("S",@sdbuffer,RamAddress,512)                   ' move to ram
      RamAddress += 256                                 ' increment ram address by 256 (half 512 as using words)
    remainder :=  sizebytes & %00000000_00000000_00000001_11111111 ' remainder (if not a multiple of 512) 
    if remainder <> 0
      fat.readdata(@sdbuffer,remainder)
      docmd("S",@sdbuffer,RamAddress,remainder)             ' send out the remainder bytes
    fat.closefile                                       ' close the file
    result := FileNumber                                ' return this file
    FileNumber +=1                                      ' point to next location

PUB FetchRamBuffer
    result := @rambuffer    

PRI NextLocation                                         ' get the next free location in ram based on last files size
    if FileNumber == 0
       RamLocation[0] := RamDiskStart                              ' work out start of this file based on last file size
    else   
       RamLocation[FileNumber] := RamLocation[FileNumber - 1] + RamSize[FileNumber - 1]  ' get location in ram
    result := RamLocation[FileNumber]
    
PUB SDBMPtoRam(stringptr) | width,height,w,i,ramaddress       ' .bmp from sd to ram and convert to 2 byte ili format and reverse order along the way
' http://en.wikipedia.org/wiki/BMP_file_format scroll down about 1/3 of the way for the header formats
' get the width, height
' store these as 2 longs at the beginning of the ram file
' the bitmap format starts at the last row and works up so need to reverse the order
    ramaddress := NextLocation                          ' get the next ram location
    OpenFileRead(stringptr)
    fat.readdata(@sdbuffer,$36)                         ' get the header 0x36 hex bytes
    width := sdbuffer[$12] + sdbuffer[$13] << 8         ' only read in two bytes as never will be >64k wide or high
    height := sdbuffer[$16] + sdbuffer[$17] << 8
    w := width * 3                                      ' this number of bytes per row, and round up to nearest 4 bytes et 327 goes to 328 The size of each row is rounded up to a multiple of 4 bytes (a 32-bit DWORD) by padding.
    w +=3                                               ' add 3 and
    w &= %11111111_11111111_11111111_11111100           ' round down
    docmd("S",@sdbuffer,ramaddress,$36)                 ' store header to ram
    i := $36 + ramaddress + ((height - 1) * width)  ' start + header and on the last row and work back
    repeat height
      fat.readdata(@sdbuffer,w)                         ' read in the first row
      docmd("F",@sdbuffer,@rambuffer,width)
      HubToRam(i,width)                                 ' send to ram
      i -= width                                        ' subtract the width, move up the picture
    fat.closefile
    Ramsize[Filenumber] := $36 + (width * height)      ' size of bitmap file in words
    result := FileNumber                                ' returns the current file number
    FileNumber += 1                                     ' add 1 to filenumber

PUB DrawBMPRam(f,x,y) | rampointer,width,height          ' pass the filenumber and the x and y location, searches for this file and if found, displays it
   rampointer := RamLocation[f]                         ' find the start of the file
   docmd("T",@sdbuffer,rampointer,$36)                      ' read in the header
   width := sdbuffer[$12] + sdbuffer[$13] << 8         ' only read in two bytes as never will be >64k wide or high
   height := sdbuffer[$16] + sdbuffer[$17] << 8
   Draw(x,y,x+width-1,y+height-1)                    ' set up the area to draw
   RamToDisplay(RamPointer + $36,width*height)       ' skip the header and display
   result := (width <<16) | (height & $0000FFFF)     ' return the width and height combined into a long

' ***************** end ram disk routines    

PUB MergeBackgroundIconRam(stringptr,x,y) | i,j,k,n                                 ' merge icon and background - if mask =0 then background, else icon (not classic alpha compositing but close enough)   
     ' i is background counter, j is icon counter, k = mask counter
    ' assumes desktop is in ramfile 0, mask in ramfile 1, icon in ramfile 2
    FileNumber := 2 
    SDBMPtoRam(stringptr)                               ' load icon into ram  
    Filenumber := 2                                     ' reset filenumber
    i := RamLocation[0] + $36 + (y*(screenwidth+1)+x)   ' location of picture pixel in ram
    j := RamLocation[2] + $36                           ' location of icon in pixel in ram
    k := RamLocation[1] + $36                           ' location of mask pixel in ram
    repeat 60
      docmd("T",@sdbuffer,i,59)                             ' get 59 pixels from background
      docmd("T",@rambuffer,j,59)                            ' get 59 icon pixels
      docmd("T",@buffer2,k,59)                              ' get words for the mask
      'repeat n from 0 to 58
      '  if (word[@buffer2][n] & %00000000_00011111) > %10000 ' RGB are the same, so use BBBBB, mask 5 low bits, and if > %10000
      '     word[@sdbuffer][n] :=word[@rambuffer][n]     ' if >128 then replace with icon pixel
      docmd("X",@sdbuffer,@rambuffer,@buffer2)              ' pasm version of the above 3 lines    
      docmd("S",@sdbuffer,i,59)                             ' send back modified line to the picture buffer    
      i += screenwidth+1                              ' next line in picture
      j +=59                                            ' icon next line
      k +=59                                            ' next line in mask (bytes) - buffer2     



PUB OpenFileRead(stringptr) ' open a file for reading and display a message if it fails
   result := \fat.openfile(stringptr,"R")
   if fat.partitionerror <> 0                           ' failed to find the file
     Text(stringptr)                         ' print the filename
     Text(result)                            ' print the error message
     repeat                                             ' stop the program

PRI TouchValue | xval,yval,x,y ' returns % values 0-100% (or 255 for no touch)
      xval := TouchX
      yval := TouchY
      if (xval => 500) and (yval =< 3800)                       ' both have to be valid for a touch
        pause1ms(2)                               ' tiny delay then read again as the first reads may not have the correct value
        xval := TouchX
        yval := TouchY
        if (xval => 500) and (yval =< 3800)                 ' and if valid again, now more likely this value is correct
            x := xval - 440                                    ' scale 0-3360
            x /=  32                                     ' scale 0-100
            x <#= 100                                    ' max 100
            x #>=0                                       ' min 0
            x := 100 - x                                 ' so left edge on left
            y := yval -420                               ' scale 0-3180
            y /= 32                                      ' scale 0-100
            y <#= 100                                    ' max 100
            y #>=0                                       ' min 0
            y := 100-y                                   ' so 0 is top left     
            result := (y <<8) | x                       ' return 00000000_00000000_YYYYYYYY_XXXXXXXX   
      if (xval < 500) or (xval >3800)
        result := $0000FFFF                             ' invalid number(s)

PUB TouchXPercent
    result := TouchValue & 255                          ' decode xval 0-100%

PUB TouchYPercent
    result := TouchValue >> 8                           ' decode yval 0-100%  

PUB TouchSwishX | xval,oldxval,difference,average ' returns result = +ve or -ve depending on swish direction     ' did finger move left or right?
      SelectSPIGroup ' turn on the spi touchscreen
      oldxval := 255' for startup
      average := 0
      repeat until (average > 30) or (average < -30)
        ' possibly add a decay function here for average
        pause1ms(50)
        xval := TouchXPercent                        ' decode xval 0-100%
        if xval == 255                                  ' not touched
          oldxval := 255                                ' reset oldxval as well
        else
          if oldxval <>255                                ' discard the first reading as the screen is touched
            difference := xval - oldxval
            'printdecimal(xval)
            'printdecimal(difference)
            average := average + difference             ' 
            'printdecimal(average)
            'crlf
          oldxval := xval                                 ' store for next time
      result := average                                 ' return negative or positive value

PUB TouchSwishY | yval,oldyval,difference,average ' returns result = +ve or -ve depending on swish direction     ' did finger move left or right?
      SelectSPIGroup ' turn on the spi touchscreen
      oldyval := 255' for startup
      average := 0
      repeat until (average > 30) or (average < -30)
        ' possibly add a decay function here for average
        pause1ms(50)
        yval := TouchYPercent                        ' decode xval 0-100%
        if yval == 255                                  ' not touched
          oldyval := 255                                ' reset oldyval as well
        else
          if oldyval <>255                                ' discard the first reading as the screen is touched
            difference := yval - oldyval
            average := average + difference             ' 
          oldyval := yval                                 ' store for next time
      result := average                                 ' return negative or positive value             

PUB TouchStart
      SPI.start(3,0)                                        ' delay,state, start clock high

PRI TouchX
      SPI.SHIFTOUT(Touch_DataIn, Touch_Clock, 5, 8 , %1101_0000)  ' reads x from  500 to 3700  (off < 500 )
      result := SPI.SHIFTIN(Touch_DataOut,Touch_Clock,2, 12)

PRI TouchY      
      SPI.SHIFTOUT(Touch_DataIn, Touch_Clock, 5, 8 , %1001_0000)  ' reads y from 400 to 3800 (off > 3800 ) 
      result := SPI.SHIFTIN(Touch_DataOut,Touch_Clock,2, 12)  

PUB TextBox(x1,y1,x2,y2)                           ' draw the outline of the box
    ClearRectangle(x1,y1,x2,y2,RGBtoWord(255,255,255))
    Drawline(x1-1,y1-1,x2,y1-1,RGBtoWord(64,64,64))        ' gray outline lines - extract values from paint shop pro
    Drawline(x1-2,y1-2,x2+1,y1-2,RGBtoWord(128,128,128))   ' horizontal upper light gray   
    Drawline(x1-1,y1-1,x1-1,y2,RGBtoWord(64,64,64))         ' vertical left gray
    Drawline(x1-2,y1-2,x1-2,y2+1,RGBtoWord(128,128,128))    ' vertical left light gray
    Drawline(x2+2,y1-2,x2+2,y2+2,RGBtoWord(255,255,255))    ' vertical right white
    Drawline(x1-2,y2+2,x2+2,y2+2,RGBtoWord(255,255,255))   ' horizontal lower white

PUB Selectbox(x1,y1,x2,y2)
   Textbox(x1,y1,x2,y2)
   VerticalScroll(x2,y1,y2)                             ' draw a vertical scroll bar

PUB VerticalScroll(x2,y1,y2)
   Arrowbox(x2-15,y1,15)                                   ' upper arrow box background
   Arrowbox(x2-15,y2-15,15)                                ' lower arrow box background
   UpArrow(x2,y1)                                          ' arrow
   DownArrow(x2,y2)                                        ' arrow

PUB MoveVerticalScroll(x2,y1,y2,yvalue)                    ' pass yvalue
   Drawline(x2-15,y1+16,x2,y2-16,RGBtoWord(237,232,228))   ' windows crosshatches 212,208,200 but we can send the exact gray which is simpler than crosshatching
   if yvalue < (y1+15)                                    ' upper range
     yvalue := y1+15
   if yvalue > (y2-15-11)                                    ' lower range 15 for arrow and slider height
     yvalue := y2-15-11                                   ' 15 for down arrow 10 for slider height
  Arrowbox(x2-15,yvalue,10)                               ' slider is 10 high
  

PRI UpArrow(x1,y1)
   Drawline(x1-7,y1+6,x1-7,y1+6,0)                      ' black pixel
   Drawline(x1-8,y1+7,x1-6,y1+7,0)                      ' lines
   Drawline(x1-9,y1+8,x1-5,y1+8,0)                      ' lines
   Drawline(x1-10,y1+9,x1-4,y1+9,0)                     ' lines

PRI DownArrow(x1,y1)
   Drawline(x1-7,y1-6,x1-7,y1-6,0)                      ' black pixel
   Drawline(x1-8,y1-7,x1-6,y1-7,0)                      ' lines
   Drawline(x1-9,y1-8,x1-5,y1-8,0)                      ' lines
   Drawline(x1-10,y1-9,x1-4,y1-9,0)                     ' lines     

PRI ArrowBox(x1,y1,h)                                   ' h is height
   Drawline(x1,y1,x1+14,y1+h-1,RGBtoWord(212,208,200))   ' gray upper arrow background
   Drawline(x1+15,y1,x1+15,y1+h,RGBtoWord(64,64,64))   ' vertical left gray   
   Drawline(x1,y1+h,x1+15,y1+h,RGBtoWord(64,64,64))   ' horizontal lower gray
   Drawline(x1+1,y1+1,x1+13,y1+1,RGBtoWord(255,255,255))' upper white line
   Drawline(x1+1,y1+1,x1+1,y1+h-2,RGBtoWord(255,255,255))' horizontal white line
   Drawline(x1+1,y1+h-1,x1+14,y1+h-1,RGBtoWord(128,128,128))' horizontal dark gray
   Drawline(x1+14,y1+1,x1+14,y1+h-1,RGBtoWord(128,128,128))' vertical gray
   
PUB DrawCheckbox(x,y,Checked,stringptr,c1,c2,f) | i                ' draw a checkbox x,y,checked, string, file for checked, file for unchecked, font file
   if checked
     DrawBMPRam(c1,x,y) 
   else
     DrawBMPRam(c2,x,y)  
   curx := x+20                                         ' add 20 width
   cury := y
   TextT(f,stringptr)

PUB DrawRadioBox(x,y,Checked,Stringptr,c1,c2,f)| i              ' draw a radio box
   if checked
     DrawBMPRam(c1,x,y)
   else
     DrawBMPRam(c2,x,y) 
   curx := x+20                                         ' add 20 width
   cury := y
   TextT(f,stringptr)
                                           
PUB DrawButton(x,y,Stringptr,c1,f)                ' draw a radio box
   DrawBMPRam(c1,x,y) 
   curx := x + (84/2)-(str.len(stringptr)*4 / 2)        ' centre, avg width 4 per char 84 could be a variable width passed to this routine
   cury := y + 8
   TextT(f,stringptr)     

PUB TextChar(FileN,ascii,x,y) | jump,size,width,height,ramaddress,xoffset,yoffset,xadvance   ' read out a character from a .ifn file stored in ram
  ' pass Fonttable = global
  ' moves to next line if off the end
    ramaddress := RamLocation[FileN]
    jump := ReadRamLong(ramaddress,(ascii << 2) + 256)   ' jump location = ascii *4 plus fonttable + 256
    docmd("T",@buffer2,ramaddress+jump>>1,20)                ' read in the width height as a block = quicker than readramlong
    size := long[@buffer2][0]                           ' size in pixels of this character
    xadvance := long[@buffer2][5]                       ' amount to move to next character   
    'if size > 0 and ascii <> 32                         ' no need to print anything if size is zero or read more font data
    width := long[@buffer2][1]                        ' width
    height := long[@buffer2][2]                       ' height
    xoffset := long[@buffer2][3]                      ' xoffset to move
    yoffset := long[@buffer2][4]                      ' yoffset to move
    if (x+width -1 + xoffset) > screenwidth
        crlf                                            ' do a new line if it won't fit
        x := curx
        y := cury
    if ascii < 32
      xadvance := 0                                     ' add nothing if a non displaying character
    if ascii == 13                                  ' carriage return
        cr
        x := curx
    if ascii ==10                                   ' line feed (new line)
        lf
        y := cury
    if ascii > 32 ' space not displayed
      Draw(x+xoffset,y+yoffset,x+width-1+xoffset,y+height-1+yoffset)      ' draw on screen
      RamToDisplay(ramaddress+((jump+32)>>1),size)       ' move bytes from ram out to the display
    return x + xadvance    

PUB crlf
    cr
    lf

PUB cr                                                  ' carriage return, to left margin
    curx := Margin                                      ' some characters are -1 xoffset so won't display

PUB lf                                                  ' line feed, move up one
    cury += FontHeight
    if cury > screenheight
      ClearRectangle(0,0,screenwidth,screenheight,BackFontColor) 
      cury := 0                                         ' if off the bottom of the screen then clear screen start at the top (VT00 does scrolling)
    ' don't clear background here, do in calling routine
    'ClearFontBackground(0,cury,screenwidth,cury + FontHeight) ' clear background line to the background color ready to draw          

PUB TextT(FileN,stringptr)                 'print at curx,cury ising file number in ramdisk
'' Print a zero-terminated string
  repeat strsize(stringptr)
    curx := TextChar(FileN,byte[stringptr++],curx,cury)    


PUB SetCursor(x,y)               
    curx := x
    cury := y


PUB ReadRamLong(WordStart,Bytevalue)                   ' easier calculating
    docmd("T",@result,Wordstart+Bytevalue>>1,2)
    return result

PUB LoadFont(stringptr) | filesize,i                ' read a .ifn file premade in angelcode bmfont and then the vb.net program
    SDtoRam(stringptr)                                  ' font to ram disk, adds one to filenumber
    docmd("T",@rambuffer,RamLocation[Filenumber-1],128) ' read first 128 words =256 bytes back
    FontHeight := byte[@rambuffer][251]                 ' font height
    repeat i from 0 to 2
      sdbuffer[i] := byte[@rambuffer][35+i]
    docmd("E",@sdbuffer,@buffer2,1)                         ' convert to 2 byte .ili format
    BackFontColor := ( byte[@buffer2][1] << 8) | byte[@buffer2][0] ' correct order
    result := FileNumber-1                              ' return pointer to the font file in ram

PUB HubToRam(RamAddress,Number)                         ' send pixels from rambuffer hub to ram
   docmd("S",@rambuffer,RamAddress,Number)

PUB RamToHub(RamAddress,Number)                         ' send pixels from ram to hub at rambuffer
   docmd("T",@rambuffer,RamAddress,Number)

PUB RamToDisplay(RamAddress,Number)                     ' send pixels from ram to display
   'if SSD1289 == true
   '  EnableDual                                         ' for the SSD1289 need to disable the display after sending any data
   docmd("U",0,RamAddress,Number)
   'if SSD1289 == true
   '  DisableDual                                        ' if SSD1289 then disable displays

PUB HubToDisplay(HubAddress,Number)                     ' send pixels from hub to display
   'if SSD1289 == true
   '  EnableDual                                         ' for the SSD1289 need to disable the display after sending any data
   docmd("V",HubAddress,0,Number)
   'if SSD1289 == true
   '  DisableDual                                         ' if SSD1289 then disable displays   

PUB Change137(group)
   docmd("Y",group,0,0)                                     ' change the current group

PUB Change161(RamAddress)                                ' pass ramaddress. Returns in group 1
    docmd("Z",0,RamAddress,0)       

PUB Clearscreen(color)                                  ' uses 2 byte ILI color (see below for rgb conversion)
    wordfill(@rambuffer,color,256)
    Draw(0,0,screenwidth,screenheight)                           ' set up the area of the screen to draw in   
    repeat 300
      HubToDisplay(@rambuffer,256)

PUB RGBtoWord(R,G,B)                                     ' convert RGB to 2 byte ILI color 
    sdbuffer[0] := R
    sdbuffer[1] := G
    sdbuffer[2] := B                        
    docmd("E",@sdbuffer,@buffer2,1)
    result := ( byte[@buffer2][1] << 8) | byte[@buffer2][0] ' correct order 
    
PUB ClearRectangle(x1,y1,x2,y2,ColorWord) | i,size,remainder ' clears an area of the screen to the current font background color
    size := (x2-x1+1) * (y2 - y1 +1)                    ' number of pixels
    remainder := size & 255                             ' if not a whole number of 256 pixels
    repeat i from 0 to 255                              ' move the background font colour to the buffer
      rambuffer[i] := ColorWord
    Draw(x1,y1,x2,y2)                                   ' set up the area of the screen to draw in
    repeat size >> 8                                    ' 256 pixel blocks
      docmd("V",@rambuffer,0,256)                    
    if remainder <> 0  
      docmd("V",@rambuffer,0,remainder)                     ' do the remainder

PUB Drawline(x1,y1,x2,y2,ColorWord)                     ' draws a vertical or horizontal line
   ClearRectangle(x1,y1,x2,y2,ColorWord)                ' same as drawing a rectangle 
      
PUB SetOrientation(e) ' change orientation true = portrait. Changes global variable 'orientation' and also screen width and height
   orientation := e
   ChangeOrientation(orientation)
   if orientation
     screenwidth :=239
     screenheight :=319
   else
     screenwidth :=319
     screenheight :=239

PUB ChangeOrientation(n) ' pass true = portrait or false = landscape, changes global variable orientation in this object
' command $0001
'8.2.4. Driver Output Control (R01h)
'R/W RS D15 D14 D13 D12 D11 D10 D9 D8 D7 D6 D5 D4 D3 D2 D1 D0
'W 1 0 0 0 0 0 SM 0 SS 0 0 0 0 0 0 0 0
'SS: Select the shift direction of outputs from the source driver.
'When SS = 0, the shift direction of outputs is from S1 to S720
'When SS = 1, the shift direction of outputs is from S720 to S1.
'In addition to the shift direction, the settings for both SS and BGR bits are required to change the
'assignment of R, G, B dots to the source driver pins.
'To assign R, G, B dots to the source driver pins from S1 to S720, set SS = 0.
'To assign R, G, B dots to the source driver pins from S720 to S1, set SS = 1.

' command $0003
 'R/W RS D15 D14 D13 D12 D11 D10 D9  D8 D7  D6   D5   D4   D3 D2 D1 D0
   'W   1  TRI DFM 0   BGR 0   0   HWM 0  ORG 0    I/D1 I/D0 AM 0  0  0
   ' When AM = 0, the address is updated in horizontal writing direction.
   'When AM = 1, the address is updated in vertical writing direction.
   'I/D[1:0] Control the address counter (AC) to automatically increase or decrease by 1 when update one pixel
   ' display data.
    SelectMemGroup                                      ' 137 and appropriate pins
    orientation := n
    if displaymode == 0 or displaymode == 2                                          ' ili9325
      if orientation
        ' for origin top left and portrait mode
        Displaycmd($0001,%00000001_00000000)                           ' $0001,$0100 set SS and SM bit 0001 0100   portrait
        Displaycmd($0003,%00010000_00110000)                           ' $0003,$1030 set GRAM write direction and BGR=1. $0003 $1030  
      else
        ' for origin top right and landscape mode 
        Displaycmd($0001,%00000000_00000000)                           ' set SS and SM bit 0001 0000   landscape   
        Displaycmd($0003,%00010000_00111000)              ' landscape  $1028 = original but 1038 is correct - not mirror image


    if displaymode ==1 or displaymode == 3              ' SSD1289
      if orientation
            ' for origin top left and portrait mode
            Displaycmd($0001,$2B3F)                           ' $0001,$0100 set SS and SM bit 0001 0100   portrait
            Displaycmd($0011,$6070) '6070                          ' $0003,$1030 set GRAM write direction and BGR=1. $0003 $1030  
      else
            ' for origin top right and landscape mode 
            Displaycmd($0001,$4B3F)                           ' set SS and SM bit 0001 0000   landscape   
            Displaycmd($0011,$6838)                     ' landscape  $1028 = original but 1038 is correct - not mirror image



PUB SelectMemGroup                                      ' select group 2 for memory transfer
   spi.stop                                             ' and stop any other cogs here too if needed
   
   OUTA |= %00000000_00011111_11111111_11111111         ' set P0-P20 high 
   DIRA |= %00000000_00011111_11111111_11111111         ' set P0-P20 as outputs 
   Change137(1)
   OUTA |= %00000000_00011111_00000000_00000000         ' set P16-P20 high 
   DIRA |= %00000000_00011111_11111111_11111111         ' set P0-P20 as outputs

    
   
PUB SelectSPIGroup                                      ' select group 0 for the SPI
   OUTA |= %00000000_00000000_00000010_00000000   ' keep p9 high so doesn't reset
   DIRA |= %00000000_00000000_00000010_00000000   ' P0-P2 are used to load the 137
   Change137(2)
   OUTA |= %00000000_00000000_00000010_00000000   ' keep p9 high so doesn't reset
   DIRA |= %00000000_00000000_00000010_00000000   ' enable these pins for output
   DIRA &= %11111111_11111111_11111111_11111000   ' so no clashes with the cog using P0-P2 set these as inputs      
   TouchStart                                     ' start the touchscreen cog

PUB Text(stringptr)                          'print at curx,cury
  repeat strsize(stringptr)
    Propfont_out(byte[stringptr++])
  PropFontcrlf

PUB PropFontcrlf
    curx := 0
    cury += 32                                          ' new line at end of string
    if cury >319                                        ' bottom of screen so new screen
      curx:=0
      cury:=0

PUB Propfont_out(ascii) | address,pixels
    Draw(curx,cury,curx+15,cury+31)                     ' location to start drawing 
    address := $8000 + (ascii >> 1) << 7                ' get rom address
    repeat 32                                           ' 32 rows per character, split in two parts
      pixels := long[address]                           ' get rom font data
      pixels := pixels >> (ascii & 1)                   ' shift for odd characters
      repeat 16 ' 16 columns
        if pixels & 1
          Pixel(%00000111_11100000)                     ' foreground color RRRRRGGG_GGGBBBBB
        else
          Pixel(%00000000_00000000)                     ' background color 
        pixels := pixels >> 2                           ' alternate pixels interleaved so shift 2
      address += 4 
    curx +=16
    if curx >239                                        ' new line
      Propfontcrlf

PUB Start_ILI9325(n)
    displaymode := n                                    ' pass the displaymode
    if displaymode == 0 or displaymode == 1             ' reset one display
      ILI_Reset_High
    if displaymode == 2 or displaymode == 3             ' reset two displays
      DualResetHigh
    pause1ms(5)
    if displaymode == 0 or displaymode == 1             ' reset one display   
      ILI_Reset_Low
    if displaymode == 2 or displaymode == 3             ' reset two displays     
      DualResetLow
    pause1ms(5)
    if displaymode == 0 or displaymode == 1 
      ILI_Reset_High                                      ' returns in group 2, ie CS is high
    if displaymode == 2 or displaymode == 3  
      DualResetHigh

    if displaymode == 0 or displaymode == 1             ' one display    
      Change137(2)                                        ' replaces ILI_CS high - select another group
      ILI_RD_High
      ILI_WR_High
      pause1ms(20)
       'ILI_CS_Low
       SelectMemGroup                                      ' enables correct pins for output and sets CS low

    if displaymode == 2 or displaymode == 3             ' dual displays
      DisableDual                                       ' CS1 and CS2 high
      ILI_RD_High
      ILI_WR_High
      pause1ms(20)
      EnableDual                                        ' CS1 and CS2 low
      SelectMemGroup                                    ' and select the mem group

    if displaymode == 0 or displaymode == 2               ' ILI9325 single or dual
      ILI9325initiate
    if displaymode == 1 or displaymode == 3
      SSD1289initiate  

PRI ILI9325initiate    
          ' ************* Start Initial Sequence **********
    Displaycmd($00E5,$78F0)                             ' set SRAM internal timing
    Displaycmd($0001,$0100)                             ' set SS and SM bit 0001 0100   portrait
    Displaycmd($0002,$0700)                             ' set 1 line inversion 
    Displaycmd($0003,$1030)                             ' set GRAM write direction and BGR=1. $0003 $1030
    Displaycmd($0004,$0000)                             ' Resize register 
    Displaycmd($0008,$0207)                             ' set the back porch and front porch 
    Displaycmd($0009,$0000)                             ' set non-display area refresh cycle ISC[3:0]  
    Displaycmd($000A,$0000)                             ' FMARK function  
    Displaycmd($000C,$0000)                             ' RGB interface setting   
    Displaycmd($000D,$0000)                             ' Frame marker Position   
    Displaycmd($000F,$0000)                             ' RGB interface polarity  
'        *************Power On sequence ****************//
    Displaycmd($0010,$0000)                             ' SAP, BT[3:0], AP, DSTB, SLP, STB 
    Displaycmd($0011,$0007)                             ' DC1[2:0], DC0[2:0], VC[2:0] 
    Displaycmd($0012,$0000)                             ' VREG1OUT voltage   
    Displaycmd($0013,$0000)                             ' VDV[4:0] for VCOM amplitude  
    Displaycmd($0007,$0001)
    pause1ms(50)                                        ' Dis-charge capacitor power voltage
    Displaycmd($0010,$1090)                             ' 1490//SAP, BT[3:0], AP, DSTB, SLP, STB   
    Displaycmd($0011,$0227)                             ' DC1[2:0], DC0[2:0], VC[2:0] 
    pause1ms(50)                                        ' delay
    Displaycmd($0012,$001F)                             ' 001C// Internal reference voltage= Vci;
    pause1ms(50)                                        ' delay   
    Displaycmd($0013,$1500)                             ' $1000//1400   Set VDV[4:0] for VCOM amplitude  1A00  
    Displaycmd($0029,$0027)                             ' $0012 //001a  Set VCM[5:0] for VCOMH  //$0025  0034  
    Displaycmd($002B,$000D)                             ' Set Frame Rate   000C       
    pause1ms(50)                                        ' delay   
    Displaycmd($0020,$0000)                             ' GRAM horizontal Address   
    Displaycmd($0021,$0000)                             ' GRAM Vertical Address    
'         ----------- Adjust the Gamma Curve ----------//
    Displaycmd($0030,$0000)
    Displaycmd($0031,$0707)
    Displaycmd($0032,$0307)
    Displaycmd($0035,$0200)
    Displaycmd($0036,$0008)
    Displaycmd($0037,$0004)
    Displaycmd($0038,$0000)
    Displaycmd($0039,$0707)
    Displaycmd($003C,$0002)
    Displaycmd($003D,$1D04)
'        ------------------ Set GRAM area ---------------//
    Displaycmd($0050,$0000)                             ' Horizontal GRAM Start Address
    Displaycmd($0051,$00EF)                             ' Horizontal GRAM End Address       
    Displaycmd($0052,$0000)                             ' Vertical GRAM Start Address    
    Displaycmd($0053,$013F)                             ' Vertical GRAM Start Address       
    Displaycmd($0060,$A700)                             ' Gate Scan Line     
    Displaycmd($0061,$0001)                             ' NDL,VLE, REV    
    Displaycmd($006A,$0000)                             ' set scrolling line      
'        -------------- Partial Display Control ---------/
    Displaycmd($0080,$0000)
    Displaycmd($0081,$0000)
    Displaycmd($0082,$0000)
    Displaycmd($0083,$0000)
    Displaycmd($0084,$0000)
    Displaycmd($0085,$0000)
'        //-------------- Panel Control -------------------//
    Displaycmd($0090,$0010)
    Displaycmd($0092,$0600)
    Displaycmd($0007,$0133)                             ' 262K color and display ON
    ChangeOrientation(true)                             ' default to portrait


PRI SSD1289initiate
{       ILIcmddelay($0000,$0001)      'Turn Oscillator on                                POR-$0000
        ILIcmddelay($0003,$A8A4)      'Power control (1)                                 POR-$6664
        ILIcmddelay($000C,$0000)      'Power control (2)                                 POR- ?
        ILIcmddelay($000D,$080C)      'Power control (3)                                 POR-$0009
        ILIcmddelay($000E,$2B00)      'Power control (4)                                 POR-$3200
        ILIcmddelay($001E,$00B0)      'Power control (5)                                 POR-$0029
        ILIcmddelay($0001,$4B3F)      'Driver output control,    *landscape?? $6B3F*     POR-$433F
        ILIcmddelay($0002,$0600)      'LCD drive AC control                              POR-$0400
        ILIcmddelay($0010,$0000)      'Sleep Mode                sleep mode off          POR-$0001
        ILIcmddelay($0011,$6070)      'Entry Mode,               *landscape? $4030*      POR-$6830  
        ILIcmddelay($0005,$0000)      'Compare Register (1)                              POR-$0000
        ILIcmddelay($0006,$0000)      'Compare Register (2)                              POR-$0000
        ILIcmddelay($0016,$EF1C)      'Horizontal Porch                                  POR-$EFC1
        ILIcmddelay($0017,$0003)      'Vertical Porch                                    POR-$0003
        ILIcmddelay($0007,$0033)      'Display Control                                   POR-$0000
        ILIcmddelay($000B,$D308)      'Frame cycle Control                               POR-$D308
        ILIcmddelay($000F,$0000)      'Gate Scan start position                          POR-$0000
        ILIcmddelay($0041,$0000)      'Vertical Scroll Control (1)                       POR-$0000
        ILIcmddelay($0042,$0000)      'Vertical Scroll Control (2)                       POR-$0000
        ILIcmddelay($0048,$0000)      'First Window Start                                POR-$0000
        ILIcmddelay($0049,$013F)      'First Window End                                  POR-$013F
        ILIcmddelay($004A,$0000)      'Second Window Start                               POR-$0000
        ILIcmddelay($004B,$0000)      'Second Window End                                 POR-$013F
        ILIcmddelay($0044,$EF00)      'Horizontal Ram Address Postion                    POR-$EF00
        ILIcmddelay($0045,$0000)      'Vertical Ram Address Start                        POR-$0000
        ILIcmddelay($0046,$013F)      'Vertical Ram Address End                          POR-$013F
        ILIcmddelay($0023,$0000)      'RAM write data mask (1)                           POR-$0000
        ILIcmddelay($0024,$0000)      'RAM write data mask (2)                           POR-$0000
        ILIcmddelay($0025,$8000)      'not in datasheet?                                     
        ILIcmddelay($004f,0)          'RAM Y address counter                             POR-$0000
        ILIcmddelay($004e,0)          'RAM X address counter                             POR-$0000
        Lcd_Write_Com($0022)
}        

' based on C code

    cmdDelay($0000,$0001)
    cmdDelay($0003,$A8A4)
    cmdDelay($000C,$0000)
    cmdDelay($000D,$080C)
    cmdDelay($000E,$2B00)
    cmdDelay($001E,$00B0)
    cmdDelay($0001,$2B3F)
    cmdDelay($0002,$0600)
    cmdDelay($0010,$0000)
    cmdDelay($0011,$6070)
    cmdDelay($0005,$0000)
    cmdDelay($0006,$0000)
    cmdDelay($0016,$EF1C)
    cmdDelay($0017,$0003)
    cmdDelay($0007,$0233)        
    cmdDelay($000B,$0000)
    cmdDelay($000F,$0000)
    cmdDelay($0041,$0000)
    cmdDelay($0042,$0000)
    cmdDelay($0048,$0000)
    cmdDelay($0049,$013F)
    cmdDelay($004A,$0000)
    cmdDelay($004B,$0000)
    cmdDelay($0044,$EF00)
    cmdDelay($0045,$0000)
    cmdDelay($0046,$013F)
    cmdDelay($0030,$0707)
    cmdDelay($0031,$0204)
    cmdDelay($0032,$0204)
    cmdDelay($0033,$0502)
    cmdDelay($0034,$0507)
    cmdDelay($0035,$0204)
    cmdDelay($0036,$0204)
    cmdDelay($0037,$0502)
    cmdDelay($003A,$0302)
    cmdDelay($003B,$0302)
    cmdDelay($0023,$0000)
    cmdDelay($0024,$0000)
    cmdDelay($0025,$8000)
    cmdDelay($004f,$0000)
    cmdDelay($004e,$0000)
    Lcd_Write_Com($0022)
    
PUB Draw(x1, y1, x2, y2)                                ' sets the pixel to x1,y1 and then fills the next (x2-x1)*(y2-y1) pixels
    SelectMemGroup                                      ' set 137 to correct value
    ifnot orientation                                   ' landscape mode so swap x and y
      result :=x1                                       ' swap x1 and y1
      x1 := y1
      y1 := result
      result := x2                                      ' swap x2 and y2
      x2 :=y2
      y2 := result

    if displaymode ==0 or displaymode == 2                            ' ILI9325 display
      Displaycmd($0050,x1)
      Displaycmd($0052,y1)
      Displaycmd($0051,x2)
      Displaycmd($0053,y2)
      Displaycmd($0020,x1)
      Displaycmd($0021,y1)
      Lcd_Write_Com($0022)
      
    if displaymode ==1 or displaymode == 3                            ' SSD1289 display
      Displaycmd($0044,(x2<<8)+x1)           
      Displaycmd($0045,y1)
      Displaycmd($0046,y2)
      Displaycmd($004e,x1)
      Displaycmd($004f,y1)
      Lcd_Write_Com($0022)



PUB Pixel(pixelcolor)               ' send out a pixel
     Lcd_Write_Data(pixelcolor)

PUB pause1ms(period) | clkcycles

'' Pause execution for period (in units of 1 ms).

  clkcycles := ((clkfreq / _1ms * period) - 4296) #> 381     ' Calculate 1 ms time unit
  waitcnt(clkcycles + cnt)                                   ' Wait for designated time

PUB hex(value, digits)   ' uses internal prop font and slow pixel drawing (for very basic debugging without an SD font)
  SelectMemGroup                               ' reselect 138 for display output for debugging  
'' Print a hexadecimal number
  propfont_out("O")
  propfont_out("x")
  value <<= (8 - digits) << 2
  repeat digits
    propfont_out(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))
  PropFontcrlf

PRI cmdDelay(c,d)
   Displaycmd(c,d)
   pause1ms(1)
                   

PRI Displaycmd(c,d) ' instruction in one method
    Lcd_Write_Com(c) ' send out a word
    Lcd_Write_Data(d)

PRI Lcd_Write_Com(ILIlong)
    'if SSD1289 == true
    '  EnableDual                                        ' CS low on displays
    ILI_RS_low
    LCD_Writ_Bus(ILIlong)
    'if SSD1289 == true
    '  DisableDual                                        ' CS high on displays
    
    'tch.docmd("W",ILIlong,0,0)  ' try moving to pasm, leaves spin and cog dira as 0 so need to reset both
    'SelectMemGroup           ' not working though. Not sure if should be ILIlong or @ILIlong. neither work
    

PRI Lcd_Write_Data(ILIlong)
    if displaymode == 1 or displaymode == 3             ' cs low for SSD1289 displays
      EnableDual                                        ' CS low on displays
    ILI_RS_High
    LCD_Writ_Bus(ILIlong)
    if displaymode == 1 or displaymode == 3 
      DisableDual                                        ' CS high on displays   

PRI LCD_Writ_Bus(ILILong)
    'ILILong &= %00000000_00000000_11111111_11111111    ' mask to a word. Not needed if care taken always to send a word value
    OUTA &= %11111111_11111111_00000000_00000000        ' set P0-P15 to zero ready to OR
    OUTA |= ILILong                                     ' merge with the word to output
    ILI_WR_Low                                          ' send out the data
    ILI_WR_High

PRI ILI_RS_Low
   OUTA &= %11111111_11111011_11111111_11111111         ' P18 low

PRI ILI_RS_High   
   OUTA |= %00000000_00000100_00000000_00000000         ' P18 high

PRI ILI_WR_Low
   OUTA &= %11111111_11101111_11111111_11111111         ' p20 low

PRI ILI_WR_High
   OUTA |= %00000000_00010000_00000000_00000000         ' p20 high

PRI ILI_RD_High                                         ' not used, set high in hardware     

'PRI ILI_CS_Low                                          
'   Change137(1)                                          ' change to group 1 
'
'PRI ILI_CS_High
'   Change137(0)                                         ' change to group 0 (any group other than 1)
   

PRI ILI_RESET_Low                                       ' reset low
   Change137(2)                                         ' group 2
   DIRA |= %00000000_00000000_00000010_00000000         ' set pin 9 as an output
   OUTA &= %11111111_11111111_11111101_11111111         ' set pin 9 low

PRI ILI_RESET_High
   Change137(2)                                         ' group 2
   DIRA |= %00000000_00000000_00000010_00000000         ' set pin 9 as an output
   OUTA |= %00000000_00000000_00000010_00000000         ' set pin 9 high

' ***** dual touchscreen routines *******

PRI LatchHigh(pin)                                      ' pass pin 0-7
   LatchStart
   Latch |= (1 << pin)                                  ' set this pin high
   LatchSend

PRI LatchLow(pin)   
   LatchStart
   Latch &= !(1 << pin)                               ' set this pin low
   LatchSend

PRI LatchStart                                          ' common routine to above 2 routines
   Change137(3)                                         ' group 3 is the latch group
   DIRA |= %00000000_00000000_00000000_11111111         ' P0-P7 are outputs   

PRI LatchSend                                           ' common routine to both above routines
   OUTA := Latch                                        ' send out
   DIRA |= %00000000_01000000_00000000_00000000         ' enable pin 22 as an output
   OUTA |= %00000000_01000000_00000000_00000000         ' set P22 high - this latches the data into the latch

PRI DualResetLow
   LatchLow(0)                                         ' resets both displays

PRI DualResetHigh
   LatchHigh(0)                                        ' reset for both displays high

PRI CS1Low
   LatchLow(1)                                          ' display 1 chip select low (needs group1 low as well)
   LatchHigh(3)                                         ' led on


PRI CS1High                                         ' display 1 chip select high
   LatchHigh(1)                                         ' cs 1 high
   LatchLow(3)                                          ' led on

PRI CS2Low                                          ' display 2 chip select low
   LatchLow(2)                                          ' needs group 1 low as well
   LatchHigh(4)                                         ' led on for debugging

PRI CS2High
   LatchHigh(2)                                         ' display 2 chip select high
   LatchLow(4)                                          ' and led off

PRI EnableDual                                          ' enable both displays
   CS1Low
   CS2Low
   Change137(1)                                         ' need to do this also for CS see the logic on the schematic with the OR gates    

PRI DisableDual                                         ' disable both displays
   CS1High
   CS2High                                              
   Change137(1)                                         ' group 1  
         
' ************ cog drivers ***************8

PUB start_ram_cog : err_
' Initialise the Drac Ram driver. No actual changes to ram as the read/write routines handle this
  command := "I"
  cog := 1 + cognew(@tbp2_start, @command)
  if cog == 0
    err_ := $FF                 ' error = no cog
  else
    repeat while command        ' driver cog sets =0 when done
    err_ := errx                ' driver cog sets =0 if no error, else xx = error code

PUB stop_ram_cog
   if cog
      cogstop(cog~ - 1)      

PUB DoCmd(command_, hub_address, ram_address, block_length) : err_| a,b
' Do the command: A-Z (I is reserved for Initialise)
  'dira := 0                     ' tristate all pins with the spin dira
  a := dira                     ' store the state of these and restore at the end 
  b := outa                     
  DIRA &= %11111111_10100000_00000000_00000000 ' tristate all common pins that cog can change so no conflicts
  hubaddrs := hub_address       ' hub address start
  ramaddrs := ram_address       ' ram address start
  blocklen := block_length      ' block length
  command  := command_          ' must be last !!
' Wait for command to complete and get status
  repeat while command          ' driver cog sets =0 when done
  err_ := errx                  ' driver cog sets =0 if no error, else xx = error code
  dira := a                     ' restore dira and outa
  outa := b

CON
'' Modified code from Cluso's triblade
'' commands to move blocks of data to the ILI9325 touchscreen display
' DoCmd(command_, hub_address, ram_address, block_length)
' I - initialise     
' S - Move data from hub to ram
' T - Move data from ram to hub
' U - Move data from ram to display
' V - Hub to display
' W - not working - writecom in pasm
' E - convert from .raw RGB to two byte ILI format RRRRRGGG_GGG_BBBBB
' F - convert from .bmp BGR format to two byte ILI format
' X - merge icon and background based on a mask
' Y - Change 137 output Returns P0-P20 and P22 in HiZ. Pass hubaddrs
' Z - Set 161 pins. Returns in group 1

VAR

' communication params(5) between cog driver code - only "command" and "errx" are modified by the driver
   long  command, hubaddrs, ramaddrs, blocklen, errx, cog ' rendezvous between spin and assembly (can be used cog to cog)
'        command  = A to Z etc =0 when operation completed by cog
'        hubaddrs = hub address for data buffer
'        ramaddrs = ram address for data
'        blocklen = ram buffer length for data transfer
'        errx     = returns =0 (false=good), else <>0 (true & error code)
'        cog      = cog no of driver (set by spin start routine)
   

DAT
'' +-----------------------------------------------------------------------------------------------+
'' | Touchblade 161 Ram Driver (with grateful acknowlegements to Cluso and Average Joe)            |
'' +-----------------------------------------------------------------------------------------------+
                        org     0
tbp2_start    ' setup the pointers to the hub command interface (saves execution time later
                                      '  +-- These instructions are overwritten as variables after start
comptr                  mov     comptr, par     ' -|  hub pointer to command                
hubptr                  mov     hubptr, par     '  |  hub pointer to hub address            
ramptr                  add     hubptr, #4      '  |  hub pointer to ram address            
lenptr                  mov     ramptr, par     '  |  hub pointer to length                 
errptr                  add     ramptr, #8      '  |  hub pointer to error status           
cmd                     mov     lenptr, par     '  |  command  I/R/W/G/P/Q                  
hubaddr                 add     lenptr, #12     '  |  hub address                           
ramaddr                 mov     errptr, par     '  |  ram address                           
len                     add     errptr, #16     '  |  length                                
err                     nop                     ' -+  error status returned (=0=false=good) 


' Initialise hardware tristates everything and read/write set the pins
init                    mov     err, #0                  ' reset err=false=good
                        'mov     dira,zero                ' tristate the pins with the cog dira
                        and     dira,maskP0P20P22       ' tristates all the common pins

done                    wrlong  err, errptr             ' status  =0=false=good, else error x
                        wrlong  zero, comptr            ' command =0 (done)
' wait for a command (pause short time to reduce power)
pause
'                        mov     ctr, delay      wz      ' if =0 no pause
'              if_nz     add     ctr, cnt
'              if_nz     waitcnt ctr, #0                 ' wait for a short time (reduces power)
                        rdlong  cmd, comptr     wz      ' command ?
              if_z      jmp     #pause                  ' not yet
' decode command
                        cmp     cmd, #"S"       wz      ' hub to ram
              if_z      jmp     #pasmhubtoram           
                        cmp     cmd, #"T"       wz      ' ram to hub
              if_z      jmp     #pasmramtohub
                        cmp     cmd, #"U"       wz      ' ram to display
              if_z      jmp     #pasmramtodisplay
                        cmp     cmd, #"V"       wz      ' hub to display
              if_z      jmp     #pasmhubtodisplay           
                        cmp     cmd, #"E"       wz      ' convert 3 byte .raw format to 2 byte .ili format - hub to hub
              if_z      jmp     #rawtoiliformat
                        cmp     cmd, #"F"       wz      ' convert 3 byte .bmp format BGR to 2 byte ili format (same as E but order reversed)
              if_z      jmp     #bmptoiliformat              
 '                       cmp     cmd, #"W"       wz      ' lcdwritecom in pasm, not working
 '             if_z      jmp     #pasmlcdwritecom
                        cmp     cmd, #"X"       wz      ' merge icon and background based on a mask
              if_z      jmp     #mergeicons
                        cmp     cmd, #"Y"       wz      ' change the 137 output
              if_z      jmp     #changegroup
                        cmp     cmd, #"Z"       wz      ' set the 161 counters
              if_z      jmp     #set161          
                        cmp     cmd, #"I"       wz      ' init
              if_z      jmp     #init     
                        mov     err, cmd                ' error = cmd (unknown command)
                        jmp     #done
                        
' ----------------- common routines -------------------------------------

get_values              rdlong  hubaddr, hubptr         ' get hub address
                        rdlong  ramaddr, ramptr         ' get ram address
                        rdlong  len, lenptr             ' get length
                        mov     err, #5                 ' err=5
get_values_ret          ret

           ' Pass pasm_n = 0- 7 come to this with P0-P20 and P22 tristated and returns them as this too
set137                  or      dira,maskP22            ' pin 22 is an output
                        andn    outa,maskP22            ' set P22low so Y0-Y7 are all high
                        or      dira,maskP0P20          ' pins P0-P20 are outputs
                        and     outa,maskP0P2low        ' set these 3 pins low
                        or      outa,pasm_n             ' set the 137 pins
                        or      outa,maskP22            ' pin 22 high
set137_ret              ret                             ' return                        


load161pasm                                             ' uses ramaddr
                        or      outa,maskP0P20          ' set P0-P20 high     
                        or      dira,maskP0P20          ' output pins 0-20
                        mov     pasm_n,#0               ' group 0
                        call    #set137                 ' set the 137 output
                        and     outa,maskP0P18low       ' pins 0-18 set low
                        or      outa,ramaddr            ' output addres to 161 chips
                        or      outa,maskP19            ' clock high
                        or      outa,maskP20            ' load high
                        andn    outa,maskP19            ' clock low
                        andn    outa,maskP20            ' load low
                        or      outa,maskP19            ' clock high
                        or      outa,maskP20            ' load high 
load161pasm_ret         ret

stop                   jmp     #stop                  ' for debugging

memorytransfer          or      dira,maskP16P20         ' so /wr and other pins definitely high
                        or      outa,maskP16P20
                        mov     pasm_n,#1               ' back to group 1 for memory transfer
                        call    #set137                 ' as next routine will always be group 1
                        or      dira,maskP16P20         ' output pins 16-20
                        or      outa,maskP16P20         ' set P16-P20 high (P0-P15 set as inputs or outputs in the calling routine)
memorytransfer_ret      ret

busoutput               or      dira,maskP0P15          ' set prop pins 0-15 as outputs
busoutput_ret           ret

businput                and     dira,maskP16P31         ' set P0-P15 as inputs
businput_ret            ret  

delaynop                nop
                        nop
                        nop
                        nop
delaynop_ret            ret


' ------------------ single letter commands  -------------------------------------
 
' command S
pasmhubtoram            call    #get_values             ' get hubaddr,ramaddr,len
                        call    #load161pasm            ' load the 161 counters with ramaddr
                        call    #memorytransfer         ' set to group 1, enable P16-P20 as outputs and set P16-P20 high
                        call    #busoutput              ' set prop pins P0-P15 as outputs
hubtoram_loop           and     outa,maskP16P31         '%11111111_11111111_00000000_00000000       ' clear for output                   
                        rdword  data_16,hubaddr         ' get the word from hub
                        and     data_16,maskP0P15       ' mask to a word only
                        or      outa,data_16            ' send out the byte to P0-P15
                        andn    outa,maskP17            ' set mem write low
                        add     hubaddr,#2              ' increment by 2 bytes = 1 word. Put this here for small delay while writes
                        or      outa,maskP17            ' mem write high
                        andn    outa,maskP19            ' clock 161 low
                        or      outa,maskP19            ' clock 161 high
                        djnz    len,#hubtoram_loop      ' loop this many times
                        jmp     #init                   ' tristate pins and listen for commands

' command T
pasmramtohub            call    #get_values             ' get hubaddr,ramaddr,len
                        call    #load161pasm            ' load the 161 counters with ramaddr
                        call    #memorytransfer         ' set to group 1, enable P16-P20 as outputs and set P16-P20 high                             
                        call    #businput               ' set prop pins P0-P15 as inputs
                        andn    outa,maskP16            ' memory /rd low
ramtohub_loop           mov     data_16,ina             ' get the data
                        wrword  data_16,hubaddr         ' move data to hub
                        andn    outa,maskP19            ' clock 161 low
                        or      outa,maskP19            ' clock 161 high
                        add     hubaddr,#2              ' increment the hub address 
                        djnz    len,#ramtohub_loop
                        or      outa,maskP16            ' memory /rd high  
                        or      dira,maskP0P15          ' %00000000_00000000_11111111_11111111 restore P0-P15as outputs
                        jmp     #init                   ' ' tristate pins and listen for commands

' command U
pasmramtodisplay        call    #get_values             ' get hubaddr,ramaddr,len
                        call    #load161pasm            ' load the 161 counters with ramaddr
                        call    #memorytransfer         ' set to group 1, enable P16-P20 as outputs and set P16-P20 high                             
                        call    #businput               ' set prop pins 0-15 as inputs so doesn't interfere with the transfer
                        or      outa,maskP18            ' ILI_RS high
                        andn    outa,maskP16            ' memory /rd low  
ramtodisplay_loop       andn    outa,maskP20            ' ILI write low
                        or      outa,maskP20            ' ILI write high
                        andn    outa,maskP19            ' clock 161 low
                        or      outa,maskP19            ' clock 161 high
                        djnz    len,#ramtodisplay_loop
                        or      outa,maskP16            ' memory /rd high  
                        or      dira,maskP0P15          ' %00000000_00000000_11111111_11111111 restore P0-P15as outputs
                        jmp     #init

' command V
pasmhubtodisplay        call    #get_values             ' get hubaddr,ramaddr,len
                        call    #memorytransfer         ' set to group 1, enable P16-P20 as outputs and set P16-P20 high                             
                        call    #busoutput              ' set P0-P15 as outputs
hubtodisplay_loop       and     outa,maskP16P31         '%11111111_11111111_00000000_00000000       ' clear for output                   
                        rdword  data_16,hubaddr         ' get the word from hub
                        and     data_16,maskP0P15       ' mask to a word only
                        or      outa,data_16            ' send out the byte to P0-P15
                        andn    outa,maskP20            ' ILI write low
                        or      outa,maskP20            ' ILI write high
                        add     hubaddr,#2              ' one word
                        djnz    len,#hubtodisplay_loop
                        jmp     #init

'command E
RawtoILIformat          ' takes a .raw 3 byte RRRRRRRR GGGGGGGG BBBBBBBB and converts to 2 byte RRRRRGGG GGGBBBBB
                        ' pass hubaddr, ramaddr and len
                        ' hubaddr is source location, len is number of pixels
                        ' ramaddr is destination in hub (messy naming) and length is 2/3 of blocklength
                        call    #get_values ' gets hubaddress, ramaddress and len (ignores ramaddress)
rawloop
                        rdbyte red,hubaddr
                        add hubaddr,#1
                        rdbyte green,hubaddr
                        add hubaddr,#1
                        rdbyte blue,hubaddr
                        add hubaddr,#1
                        call #rgbtoili
                        wrbyte ililow,ramaddr
                        add ramaddr,#1
                        wrbyte ilihigh,ramaddr
                        add ramaddr,#1
                        djnz    len,#rawloop            ' loop until done 
                        jmp     #init                   ' set pins to tristate

RGBtoILI                ' pass red,green, blue, returns ililow and ilihigh
                        shr     red,#3                  ' 000RRRRR 
                        shl     red,#3                  ' RRRRR000 
                        shr     green,#2                ' 00GGGGGG
                        mov     ilihigh,green           ' ilihigh = 00GGGGGG
                        shr     ilihigh,#3              ' ilihigh = 00000GGG
                        or      ilihigh,red             ' ilihigh = RRRRRGGG
                        and     green,#%00000111        ' 00000GGG
                        shl     green,#5                ' GGG00000
                        mov     ililow,green            ' ililow = GGG00000
                        shr     blue,#3                 ' blue = 000BBBBB
                        or      ililow,blue             ' ililow = GGGBBBBB
RGBtoILI_ret            ret

BMPtoILIformat          ' takes a .bmp 3 byte BBBBBBBB GGGGGGGG RRRRRRRR and converts to 2 byte RRRRRGGG GGGBBBBB
                        ' same as E above but BGR instead of RGB
                        ' pass hubaddr, ramaddr and len
                        ' hubaddr is source location, len is number of pixels
                        ' ramaddr is destination in hub (messy naming) and length is 2/3 of blocklength
                        call    #get_values ' gets hubaddress, ramaddress and len (ignores ramaddress)
bmploop
                        rdbyte blue,hubaddr
                        add hubaddr,#1
                        rdbyte green,hubaddr
                        add hubaddr,#1
                        rdbyte red,hubaddr
                        add hubaddr,#1
                        call #rgbtoili
                        wrbyte ililow,ramaddr
                        add ramaddr,#1
                        wrbyte ilihigh,ramaddr
                        add ramaddr,#1
                        djnz    len,#bmploop            ' loop until done 
                        jmp     #init                   ' set pins to tristate
' **** command X *********************

MergeIcons              call    #get_values ' gets hubaddress, ramaddress,len which are used here as background,icon,mask
                        mov     pasm_n,#59               ' do a single row
mergeiconsloop          rdbyte  ililow,len                 ' reuse ililow, so this is rdword mask,maskcounter
                        and     ililow,#%11111             ' mask off low 5 bits and use just the blue as this is a grayscale bitmap
                        rdword  red,hubaddr              ' reuse red, so actually this is rdword background,backgroundcounter                        
                        cmp     ililow,#%10000   wc       ' compare if >128 (ie mid level gray)
              if_c      jmp     #mergeskip
                        rdword  green,ramaddr            ' reuse green, so this is rdword iconpixel, iconpixelcounter 
                        wrword  green,hubaddr            ' if replace, then move icon pixel to the background     
mergeskip               add     hubaddr,#2
                        add     ramaddr,#2
                        add     len,#2
                        djnz    pasm_n,#mergeiconsloop            ' loop until done 
                        jmp     #init                   'set pins to tristate 

' *** command Y **********
changegroup             call    #get_values             'gets hubaddr, ramaddr,len which are used here as background,icon,mask           
                        mov     pasm_n,hubaddr          ' pass hubaddr
                        call    #set137                 ' change the group
                        jmp     #init

' *** command Z **********
set161                  call    #get_values             ' gets ramaddr = the 161 value to set
                        call    #load161pasm            ' ramaddr to the 161 chips
                        call    #memorytransfer         ' change to group 1
                        jmp     #init                        

                        

'pasmlcdwritecom         call    #get_values             ' use hubaddr as the data
'                        or      dira,maskP0P20          ' set these pins high (pass all pins tristated)
'                        or      outa,maskP0P20          '  set pins high
'                        mov     pasm_n,#2               '  mem transfer
'                        call    #set138                 ' set the 138
'                        andn    outa,maskP18            ' P18 ILIRS low
'                        and     outa,maskP16P31         ' set P0-P15 low
'                        or      outa,hubaddr            ' send out the data
'                        andn    outa,maskP17            ' ILI write low
'                        or      outa,maskP17            ' ILI write high
'                        jmp     #init                   ' set pins to tristate  

' variables
pasm_n                  long    0                                    ' general purpose value
data_16                 long    0                                    ' general purpose value
ililow                  long    0                                    ' low data byte 
ilihigh                 long    0                                    ' high data byte 
red                     long    0                                    ' red, green blue variables
green                   long    0
blue                    long    0           

' constants
Zero                    long    %00000000_00000000_00000000_00000000 ' used in several places
maskP0P2low             long    %11111111_11111111_11111111_11111000 ' P0-P2 low
maskP0P20               long    %00000000_00011111_11111111_11111111 ' P0-P18 enabled for output plus P19,P20    
maskP0P18low            long    %11111111_11111000_00000000_00000000 ' P0-P18 low
maskP16                 long    %00000000_00000001_00000000_00000000 ' pin 16
maskP17                 long    %00000000_00000010_00000000_00000000 ' pin 17
maskP18                 long    %00000000_00000100_00000000_00000000 ' pin 18
maskP19                 long    %00000000_00001000_00000000_00000000 ' pin 19
maskP20                 long    %00000000_00010000_00000000_00000000 ' pin 20
maskP22                 long    %00000000_01000000_00000000_00000000 ' pin 22
maskP16P31              long    %11111111_11111111_00000000_00000000 ' pin 16 to pin 31
maskP0P15               long    %00000000_00000000_11111111_11111111 ' for masking words
maskP16P20              long    %00000000_00011111_00000000_00000000
maskP0P20P22            long    %11111111_10100000_00000000_00000000 ' for returning all group pins HiZ
                        fit     496
                  